"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("omni.models/utils");
const uuidv4 = require('uuid/v4');
const registry = {};
function load(name, version, rules, schemaVersion) {
    utils_1.assert(name !== void 0, `Missing rules engine name`, "RE-001");
    utils_1.assert(version !== void 0, `Missing rules engine version`, "RE-002");
    if ((registry[name] !== void 0) && (registry[name][version.toString()])) {
        return registry[name][version.toString()];
    }
    if (registry[name] === void 0) {
        registry[name] = {};
    }
    if (registry[name][version.toString()] === void 0) {
        utils_1.assert(rules !== void 0, `Missing rules for engine ${name}`, "RE-003");
        registry[name][version.toString()] = new RulesEngine(rules, name, version.toString(), schemaVersion);
    }
    return registry[name][version.toString()];
}
exports.load = load;
function getEngine(name, version) {
    utils_1.assert(name !== void 0, `Missing rules engine name`, "RE-001");
    utils_1.assert(version !== void 0, `Missing rules engine version`, "RE-002");
    utils_1.assert(registry[name] !== void 0, `No engine loaded with name ${name}`, "RE-004");
    utils_1.assert(registry[name][version] !== void 0, `Version ${version} not loaded for ${name}`, "RE-005");
    return registry[name][version];
}
exports.getEngine = getEngine;
exports.engines = {
    load,
    loadDecisionObject(ds) {
        load(ds.name || "unnamed", (ds.version && ds.version.toString()) || "1.0.0", ds.getRules(), ds.schemaVersion());
    },
    getEngine,
    registry
};
function censor(censor) {
    let i = 0;
    return (key, value) => {
        if (i !== 0 && typeof (censor) === "object" && typeof (value) === "object" && censor === value) {
            return "[Circular]";
        }
        if (i >= 100) {
            return typeof value === "function" ? undefined : `${typeof value}`;
        }
        ++i;
        return value;
    };
}
function addValueToJsonPath(json, path, defaultValue, rootName) {
    if (path === "ApplicantEmailValidation[]") {
        console.log(JSON.stringify(json), path);
    }
    const root = json;
    path = path.replace(/\[\]/g, ".?");
    const paths = rootName ? utils_1.getPath(`${rootName}.${path}`) : utils_1.getPath(`${path}`);
    let parentNode = {};
    let child = "";
    paths.forEach((path) => {
        if (path !== "?") {
            parentNode = json;
            try {
                if (json[path] === undefined) {
                    json[path] = {};
                }
            }
            catch (e) {
                console.error(`Failed to add path in ${paths}: ${JSON.stringify(root, undefined, 2)}`);
                throw e;
            }
            json = json[path];
            child = path;
        }
        else {
            if (parentNode[child][0] === undefined) {
                parentNode[child] = [];
            }
            json = parentNode[child];
            parentNode = json;
            child = json.length;
        }
    });
    if (defaultValue !== undefined) {
        if (typeof child === "number") {
            if (typeof defaultValue === "object") {
                const hasValue = parentNode.find((n) => JSON.stringify(n) === JSON.stringify(defaultValue));
                if (!hasValue) {
                    parentNode.push(defaultValue);
                }
            }
            else {
                if (parentNode.indexOf(defaultValue) > -1) {
                    parentNode.push(defaultValue);
                }
            }
        }
        else {
            parentNode[child] = defaultValue;
        }
        if (child === "?") {
            parentNode[child] = undefined;
        }
    }
    else {
        delete parentNode[child];
    }
    return root;
}
class RuleInstance {
    constructor(name, code, variables, enumerations, behaviour, ruleCode) {
        this.name = name;
        this.code = code;
        this.variables = variables;
        this.behaviour = behaviour;
        this.ruleCode = ruleCode;
        this.path = {};
        this.id = uuidv4();
        this.behaviour = behaviour || "OnlyOnceWithAllFacts";
        this.code = code;
        this.variables.forEach((variable) => {
            this.path[variable] = utils_1.getPath(variable);
        });
        this.enumerations = enumerations;
    }
    execute(rule, bomRoot) {
        if (rule.behaviour === "Disabled") {
            this.execute = (rule, bomRoot) => { return undefined; };
        }
        else {
            let result;
            const theResultIs = (value) => {
                if (rule.enumerations.length && (value !== undefined)) {
                    utils_1.assert(rule.enumerations.indexOf(value) > -1, `Invalid value for enumeration on ${rule.name}}`, "ENUM-01");
                }
                result = value;
                return value;
            };
            result = theResultIs(undefined);
            const space = " ", commaSpace = ", ", comma = ",";
            const fs = `fn=function(rule,bom){try{${this.code};if(result===undefined){return;}let s=rule.name;addValueToJsonPath(bom, 'bom.'+s, result);return result;}catch(e){console.error("Error in "+rule.name+"; Valid="+rule.valid+"; Error:"+e.message+" "+e.stack+"}"),console.error("     "+rule.usedVariables+": "+JSON.stringify(rule.usedValues)),console.error("     "+JSON.stringify(r.bomRoot)),console.error("     Code: "+rule.code),r.bomRoot.error=e}return rule};`;
            let fn;
            fn = (rule, engine) => {
                return undefined;
            };
            eval(fs);
            this.execute = fn;
            return fn(this, bomRoot);
        }
    }
}
exports.RuleInstance = RuleInstance;
function isEquivalent(a, b) {
    var aProps = Object.getOwnPropertyNames(a);
    var bProps = Object.getOwnPropertyNames(b);
    if (aProps.length != bProps.length) {
        return false;
    }
    for (var i = 0; i < aProps.length; i++) {
        var propName = aProps[i];
        if (a[propName] !== b[propName]) {
            return false;
        }
    }
    return true;
}
function onlyUnique(value, index, self) {
    return self.indexOf(value) === index;
}
function getDependencies(rules, bomRoot) {
    const values = {};
    const keys = [];
    const keyMap = {};
    rules.forEach(rule => {
        rule.variables.forEach(variable => {
            if (keys.indexOf(variable) === -1) {
                values[variable] = { value: utils_1.getValueForPath(bomRoot, variable.split('.')) };
                keys.push(variable);
            }
        });
        if (keys.indexOf(rule.name + "." + rule.ruleCode) === -1) {
            values[rule.name + "." + rule.ruleCode] = { value: utils_1.getValueForPath(bomRoot, rule.name.split('.')) };
            keys.push(rule.name + "." + rule.ruleCode);
        }
    });
    return values;
}
function missingDependencies(rule, values) {
    return rule.variables.reduce((sum, next) => sum + (values[next].value === undefined ? 1 : 0), 0);
}
const runRule = function (bomRoot, rule, values, rules) {
    const always = rule.behaviour === "AlwaysCalculate";
    const canExecute = always ? true : (missingDependencies(rule, values) === 0) && values[rule.name + '.' + rule.ruleCode].value === undefined;
    if (canExecute) {
        const value = rule.execute(rule, bomRoot);
        const old = values[rule.name + '.' + rule.ruleCode].value;
        if ((value !== undefined) && ((old === undefined) || !isEquivalent(value, old))) {
            values[rule.name + '.' + rule.ruleCode].value = value;
            if (always) {
                rules.push(rule);
            }
            return true;
        }
    }
    rules.push(rule);
    return false;
};
class RulesEngine {
    constructor(inputRules, name, version, schemaVersion, validateInputs = []) {
        this.inputRules = inputRules;
        this.name = name;
        this.version = version;
        this.schemaVersion = schemaVersion;
        this.validateInputs = validateInputs;
        this.iterations = 0;
        this.rules = [];
        this.id = uuidv4();
        this.maxIterations = inputRules.length + 1;
        inputRules.forEach((rule) => {
            this.rules.push(new RuleInstance(rule.name, rule.code, rule.variables, rule.enumerations || [], rule.behaviour || "OnlyOnceWithAllFacts", rule.ruleCode));
        });
    }
    toJSON() {
        return {
            name: this.name,
            version: this.version,
            rules: this.rules
        };
    }
    toObject() {
        return {
            name: this.name,
            version: this.version
        };
    }
    runInputValidation(bomRoot) {
        this.validateInputs.forEach((inputRule) => {
            if (utils_1.assertValueForPath(bomRoot, utils_1.getPath(`bom.${inputRule}`), undefined)) {
                eval(`throw new Error('Input "${inputRule}" is missing a value inside "${this.name}": ${JSON.stringify(bomRoot)}')`);
            }
        });
    }
    run(withBom, configuration) {
        this.iterations = 0;
        const start = new Date();
        let changed = false;
        const values = getDependencies(this.rules, withBom);
        let rules = [];
        this.rules.forEach(rule => {
            if (runRule(withBom, rule, values, rules)) {
                changed = true;
            }
        });
        while (changed) {
            if (this.iterations >= this.maxIterations) {
                console.log(`Max iterations of ${this.maxIterations} reached.`);
            }
            this.iterations++;
            changed = false;
            const runRules = [...rules];
            rules = [];
            runRules.forEach(rule => {
                if (runRule(withBom, rule, values, rules)) {
                    changed = true;
                }
            });
        }
        if (!!configuration && !!configuration.withStats) {
            const end = new Date();
            if (!withBom.engine)
                withBom.engine = {};
            const engine = {
                runtime: { version: this.version,
                    schema: this.schemaVersion,
                    executionTime: end.valueOf() - start.valueOf(),
                    runIterations: this.iterations,
                },
                tableResults: {},
                conditions: withBom._conditions
            };
            withBom.engine[this.name || "default"] = engine;
        }
        delete withBom._conditions;
        return withBom;
    }
}
exports.RulesEngine = RulesEngine;
//# sourceMappingURL=index.js.map